%---------------------------------------------------------------------------
% DDEChecker: a tool that implements validated simulation-based verification
% algorithms to verify bounded-time invariant properties of DDEs of
% the form x'(t) = f(x(t),x(t-r)).
% 
% Current version: v1.0
% Validated on platform: Matlab R2014a, 64bit-Ubuntu-Desktop-14.10
% Latest modification: on Apr 30, 2016, by Mingshuai Chen, SKLCS-ISCAS
% 
% Corresponding e-mail: {chenms@ios.ac.cn}
% Comments and bug-reports are higly appreciated.
%---------------------------------------------------------------------------

function sol = DDEChecker(dyn,dim,lags,tspan,unsafe,center,delta,h0,partition,bound)

if nargin ~= 10
    error('DDEChecker:argChk','Wrong number of input arguments');
end

%% initialization
simNum = 0;   %%number of simulations

%% partition the initial set
[A,B] = cutset(center,delta,partition,dim);

%% safety verification
tic;
flag = true;

% plot for dim = 1
% figure

%plot for dim = 2
% figure
% hold all;
% hd = ezplot('y1+y2',[-0.5,1.5]);
% set(hd,'LineColor','r');
% legend(hd,'upper bound of the unsafe set');
% xlabel('S');
% ylabel('x');


while(flag)
    init = A(1,:);
    
    disp('A new simulation starting from');
    disp(init);
    disp('with delta:');
    disp(B(1));
    
    simNum = simNum+1;
    
    n = 2;
    h = h0;
    t = [tspan(1)-h;tspan(1)];
    y = [init;init;];
    d = [0;B(1)];
    
    fprintf('n = ');
    while(t(end) < tspan(end))
        fprintf('%d.',n);

        h = h0;
        m = lags(1)/h;
        
        for i=length(t):-1:3
           if ((t(n)-lags(1)<=t(i)) && (t(n)-lags(1)>t(i-1)+1.0e-10))
               m = n-i;
               break;
           end
        end

        while(1)
            tn1 = t(n)+h;
            
            %         m = 1/h*lags;
            %         if m ~= fix(m)
            %             disp('Please set the stepsize h0 as an exact factor of the lag.');
            %             return;
            %         end
            
            yn1 = dyn(getl(y,n),getl(y,n-m))*h+getl(y,n);
            
            % constraint solving, also validated by HySAT.
            
            obj = @(p) p(1);
            
            cons1 = @(p) [-p(2);p(2)-h];
            
            % p(1):e,p(2):t,p(3):x,p(4):u,p(5):f,p(6):g
            if dim == 1
                cons2 = @(p) [(p(3)-getl(y,n,1))^2-getl(d,n)^2; ...,
                    (p(4)-getl(y,n-m,1))^2-getl(d,n-m)^2; ...,
                    (p(5)-dyn(getl(y,n),getl(y,n-m)))^2-p(1)^2; ...,
                    (p(6)-dyn(getl(y,n-m),getl(y,n-2*m)))^2-(1/h*(getl(d,n-m+1)-getl(d,n-m)))^2];
                if n-m <= 1
                    cons3 = @(p) [norm(dyn(p(3)+p(5)*p(2),p(4))-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                else
                    cons3 = @(p) [norm(dyn(p(3)+p(5)*p(2),p(4)+p(6)*p(2))-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                end 
                
            % p(1):e,p(2):t,p(3):x1,p(4):x2,p(5):u1,p(6):u2,p(7):f1,p(8):f2,p(9):g1,p(10):g2
            elseif dim == 2
                cons2 = @(p) [(p(3)-getl(y,n,1))^2+(p(4)-getl(y,n,2))^2-getl(d,n)^2; ...,
                    (p(5)-getl(y,n-m,1))^2+(p(6)-getl(y,n-m,2))^2-getl(d,n-m)^2; ...,
                    (p(7)-getl(dyn(getl(y,n),getl(y,n-m))',1))^2+(p(8)-getl(dyn(getl(y,n),getl(y,n-m))',2))^2-p(1)^2; ...,
                    (p(9)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',1))^2+(p(10)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',2))^2-(1/h*(getl(d,n-m+1)-getl(d,n-m)))^2];
                if n-m <= 1
                    cons3 = @(p) [norm(dyn([p(3)+p(7)*p(2);p(4)+p(8)*p(2)],[p(5);p(6)])-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                else
                    cons3 = @(p) [norm(dyn([p(3)+p(7)*p(2);p(4)+p(8)*p(2)],[p(5)+p(9)*p(2);p(6)+p(10)*p(2)])-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                end
                
            % p(1):e,p(2):t,p(3):x1,p(4):x2,p(5):x3,p(6):u1,p(7):u2,p(8):u3,p(9):f1,p(10):f2,p(11):f3,p(12):g1,p(13):g2,P(14):g3    
            elseif dim == 3
                cons2 = @(p) [(p(3)-getl(y,n,1))^2+(p(4)-getl(y,n,2))^2+(p(5)-getl(y,n,3))^2-getl(d,n)^2; ...,
                    (p(6)-getl(y,n-m,1))^2+(p(7)-getl(y,n-m,2))^2+(p(8)-getl(y,n-m,3))^2-getl(d,n-m)^2; ...,
                    (p(9)-getl(dyn(getl(y,n),getl(y,n-m))',1))^2+(p(10)-getl(dyn(getl(y,n),getl(y,n-m))',2))^2+(p(11)-getl(dyn(getl(y,n),getl(y,n-m))',3))^2-p(1)^2; ...,
                    (p(12)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',1))^2+(p(13)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',2))^2+(p(14)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',3))^2-(1/h*(getl(d,n-m+1)-getl(d,n-m)))^2];
                if n-m <= 1
                    cons3 = @(p) [norm(dyn([p(3)+p(9)*p(2);p(4)+p(10)*p(2);p(5)+P(11)*p(2)],[p(6);p(7);p(8)])-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                else
                    cons3 = @(p) [norm(dyn([p(3)+p(9)*p(2);p(4)+p(10)*p(2);p(5)+P(11)*p(2)],[p(6)+p(12)*p(2);p(7)+p(13)*p(2);p(8)+p(14)*p(2)])-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                end
                
            % p(1):e,p(2):t,p(3):x1,p(4):x2,p(5):x3,p(6):x4,p(7):u1,p(8):u2,p(9):u3,p(10):u4,p(11):f1,p(12):f2,p(13):f3,p(14):f4,p(15):g1,p(16):g2,p(17):g3,p(18):g4
            elseif dim ==4
                cons2 = @(p) [(p(3)-getl(y,n,1))^2+(p(4)-getl(y,n,2))^2+(p(5)-getl(y,n,3))^2+(p(6)-getl(y,n,4))^2-getl(d,n)^2; ...,
                    (p(7)-getl(y,n-m,1))^2+(p(8)-getl(y,n-m,2))^2+(p(9)-getl(y,n-m,3))^2+(p(10)-getl(y,n-m,4))^2-getl(d,n-m)^2; ...,
                    (p(11)-getl(dyn(getl(y,n),getl(y,n-m))',1))^2+(p(12)-getl(dyn(getl(y,n),getl(y,n-m))',2))^2+(p(13)-getl(dyn(getl(y,n),getl(y,n-m))',3))^2+(p(14)-getl(dyn(getl(y,n),getl(y,n-m))',4))^2-p(1)^2; ...,
                    (p(15)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',1))^2+(p(16)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',2))^2+(p(17)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',3))^2+(p(18)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',4))^2-(1/h*(getl(d,n-m+1)-getl(d,n-m)))^2];
                if n-m <= 1
                    cons3 = @(p) [norm(dyn([p(3)+p(11)*p(2);p(4)+p(12)*p(2);p(5)+p(13)*p(2);p(6)+p(14)*p(2)],[p(7);p(8);p(9);p(10)])-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                else
                    cons3 = @(p) [norm(dyn([p(3)+p(11)*p(2);p(4)+p(12)*p(2);p(5)+p(13)*p(2);p(6)+p(14)*p(2)],[p(7)+p(15)*p(2);p(8)+p(16)*p(2);p(9)+p(17)*p(2);p(10)+p(18)*p(2)])-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                end
            % p(1):e,p(2):t,p(3):x1,p(4):x2,p(5):x3,p(6):x4,p(7):x5,p(8):u1,p(9):u2,p(10):u3,p(11):u4,p(12):u5,p(13):f1,p(14):f2,p(15):f3,p(16):f4,p(17):f5,p(18):g1,p(19):g2,p(20):g3,p(21):g4,p(22):g5
            elseif dim ==5
                cons2 = @(p) [(p(3)-getl(y,n,1))^2+(p(4)-getl(y,n,2))^2+(p(5)-getl(y,n,3))^2+(p(6)-getl(y,n,4))^2+(p(7)-getl(y,n,5))^2-getl(d,n)^2; ...,
                    (p(8)-getl(y,n-m,1))^2+(p(9)-getl(y,n-m,2))^2+(p(10)-getl(y,n-m,3))^2+(p(11)-getl(y,n-m,4))^2+(p(12)-getl(y,n-m,5))^2-getl(d,n-m)^2; ...,
                    (p(13)-getl(dyn(getl(y,n),getl(y,n-m))',1))^2+(p(14)-getl(dyn(getl(y,n),getl(y,n-m))',2))^2+(p(15)-getl(dyn(getl(y,n),getl(y,n-m))',3))^2+(p(16)-getl(dyn(getl(y,n),getl(y,n-m))',4))^2+(p(17)-getl(dyn(getl(y,n),getl(y,n-m))',5))^2-p(1)^2; ...,
                    (p(18)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',1))^2+(p(19)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',2))^2+(p(20)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',3))^2+(p(21)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',4))^2+(p(22)-getl(dyn(getl(y,n-m),getl(y,n-2*m))',5))^2-(1/h*(getl(d,n-m+1)-getl(d,n-m)))^2];
                if n-m <= 1
                    cons3 = @(p) [norm(dyn([p(3)+p(13)*p(2);p(4)+p(14)*p(2);p(5)+p(15)*p(2);p(6)+p(16)*p(2);p(7)+p(17)*p(2)],[p(8);p(9);p(10);p(11);p(12)])-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                else
                    cons3 = @(p) [norm(dyn([p(3)+p(13)*p(2);p(4)+p(14)*p(2);p(5)+p(15)*p(2);p(6)+p(16)*p(2);p(7)+p(17)*p(2)],[p(8)+p(18)*p(2);p(9)+p(19)*p(2);p(10)+p(20)*p(2);p(11)+p(21)*p(2);p(12)+p(22)*p(2)])-dyn(getl(y,n),getl(y,n-m)))-p(1)];
                end
            end
            
            conseq = [];
            nonlinfcn = @(p)deal([cons1(p);cons2(p);cons3(p)],conseq);
            
            [eee,fval,exitflag,output]=fmincon(obj,[0;0.5*h;getl(y,n)';getl(y,n-m)';dyn(getl(y,n),getl(y,n-m))';dyn(getl(y,n-m),getl(y,n-2*m))'], ...,
                [],[],[],[],[],[],nonlinfcn,optimoptions('fmincon','Display','none'));
            if exitflag >= 0
                e = fval;               
                dn1 = getl(d,n)+h*e;
                break;
            else
                h = 0.5*h;
                if h < 1.0e-10
                    fprintf('\nOops, the stepsize goes too small...\n');
                    return;
                else
                    fprintf('\nThe optimizer failed to find the minimal e. Reducing the stepsize to %d\n',h);
                end
            end
            
        end
        
        t = [t;tn1];
        y = [y;yn1];
        d = [d;dn1];
        
        n = n+1;
        
    end
    fprintf('\n\n');

    %plot for dim = 1

%     subplot(2,3,simNum);
%     hd = plot(t(2:end),y(2:end),'k-',t(2:end),y(2:end)+d(2:end),'K--',t(2:end),y(2:end)-d(2:end),'k--');
%     axis([0,5.2,0.4,1.7]);
%     set(hd(1),'LineWidth',2);
%     set(hd(2),'LineWidth',1.5);
%     set(hd(3),'LineWidth',1.5);
%     hline = refline([0,unsafe(0)]);
%     set(hline,'LineWidth',2);

    %plot for dim = 2
%     curvey = plot(y(2:end,1),y(2:end,2),'g-');
%     if simNum == 1
%         legend(curvey,'numerical solution (S;x)');
%     end
%     circles(y(2:end,1),y(2:end,2),d(2:end));

    y_bound = zeros(size(y));
    y_bound_other = zeros(size(y));
%     bounds need to be balls for multi-dimensional systems, rather than cubes.
    for k = 1:dim
        if strcmp(bound,'upper')
            y_bound(:,k) = y(:,k)+d;
            y_bound_other(:,k) = y(:,k)-d;
        elseif strcmp(bound,'lower')
            y_bound(:,k) = y(:,k)-d;
            y_bound_other(:,k) = y(:,k)+d;
        end
    end
    
    if issafe(y_bound,unsafe) == 1
        disp('This trajectory is safe, go to the next...');
        fprintf('\n');
        if length(B) == 1
            flag = false;
            sol = 'SAFE';
            disp('All the partitions has been verified to be safe.');
            fprintf('\n');
        else
            A = A(2:end,:);
            B = B(2:end);
        end
    else
        if issafe(y_bound_other,unsafe) == 0
            flag = false;
            sol = 'UNSAFE';
            disp('Finding one trajectory in the unsafe set.');
            fprintf('\n');
        else
            disp('Finding one suspect reachtube, partition it...');
            fprintf('\n');
            [newA,newB] = cutset(init,B(1),2,dim);
            [A,B] = addnewpartition(A,B,newA,newB);
        end
    
    end
end
    
toc;

fprintf('Verification result: %s\n',sol);
fprintf('Number of simulations: %d\n\n',simNum);

%-----------------------------------------------------------------------
